/// Copyright (c) 2021 Razeware LLC
/// 
/// Permission is hereby granted, free of charge, to any person obtaining a copy
/// of this software and associated documentation files (the "Software"), to deal
/// in the Software without restriction, including without limitation the rights
/// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
/// copies of the Software, and to permit persons to whom the Software is
/// furnished to do so, subject to the following conditions:
/// 
/// The above copyright notice and this permission notice shall be included in
/// all copies or substantial portions of the Software.
/// 
/// Notwithstanding the foregoing, you may not use, copy, modify, merge, publish,
/// distribute, sublicense, create a derivative work, and/or sell copies of the
/// Software in any work that is designed, intended, or marketed for pedagogical or
/// instructional purposes related to programming, coding, application development,
/// or information technology.  Permission for such use, copying, modification,
/// merger, publication, distribution, sublicensing, creation of derivative works,
/// or sale is expressly withheld.
/// 
/// This project and source code may use libraries or frameworks that are
/// released under various Open-Source licenses. Use of those libraries and
/// frameworks are governed by their own individual licenses.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
/// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
/// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
/// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
/// THE SOFTWARE.

import MetalKit

// swiftlint:disable implicitly_unwrapped_optional
// swiftlint:disable force_cast

class Renderer: NSObject {
  static var device: MTLDevice!
  static var commandQueue: MTLCommandQueue!
  static var library: MTLLibrary!
  var pipelineState: MTLRenderPipelineState!

  let sphere: Sphere
  var uniforms=Uniforms()
    
  init(metalView: MTKView) {
    guard
      let device = MTLCreateSystemDefaultDevice(),
      let commandQueue = device.makeCommandQueue() else {
        fatalError("GPU not available")
    }
    Renderer.device = device
    Renderer.commandQueue = commandQueue
    metalView.device = device
      /*sphere = Sphere(device: device, latitude: 28, longitude: 28)
      let vertexDescriptor = MTLVertexDescriptor()
      vertexDescriptor.attributes[0].format = .float3
      vertexDescriptor.attributes[0].offset = 0
      vertexDescriptor.attributes[0].bufferIndex = 0
      
      vertexDescriptor.layouts[0].stride = MemoryLayout<SIMD3<Float>>.stride*/
      //sphere = Sphere(device: device, latitude: 28, longitude: 28)
    // create the shader function library
    let library = device.makeDefaultLibrary()
    let vertexFunction = library?.makeFunction(name: "vertex_sphere")
    let fragmentFunction =
      library?.makeFunction(name: "fragment_sphere")

    // create the pipeline state
    let pipelineDescriptor = MTLRenderPipelineDescriptor()
    pipelineDescriptor.vertexFunction = vertexFunction
    pipelineDescriptor.fragmentFunction = fragmentFunction
     // pipelineDescriptor.vertexDescriptor =
        //MTKMetalVertexDescriptorFromModelIO(Sphere.vertexDescriptor)
    pipelineDescriptor.colorAttachments[0].pixelFormat =
      metalView.colorPixelFormat
    do {
        pipelineState = try device.makeRenderPipelineState(descriptor: pipelineDescriptor)
    } catch let error {
      fatalError(error.localizedDescription)
    }

    // create the sphere
    sphere = Sphere(device: device, latitude: 28, longitude: 28)
    super.init()
    metalView.clearColor = MTLClearColor(
      red: 1.0,
      green: 1.0,
      blue: 0.8,
      alpha: 1.0)
    metalView.delegate = self
      
      let translation = float4x4(translation: [0, 0.3, 0])
      let rotation = float4x4(rotation: [0, Float(45).degreesToRadians, 0])
      uniforms.modelMatrix = translation * rotation
      uniforms.viewMatrix = float4x4(translation: [0.8, 0, 0]).inverse
      mtkView(metalView, drawableSizeWillChange: metalView.bounds.size)
  }
}

extension Renderer: MTKViewDelegate {
  func mtkView(_ view: MTKView,drawableSizeWillChange size: CGSize) {
      
      let aspect = Float(view.bounds.width) / Float(view.bounds.height)
      let projectionMatrix =
        float4x4(projectionFov: Float(70).degreesToRadians,
                 near: 0.001,
                 far: 100,
                 aspect: aspect)
      uniforms.projectionMatrix = projectionMatrix
  }

  func draw(in view: MTKView) {
    guard
      let commandBuffer = Renderer.commandQueue.makeCommandBuffer(),
      let descriptor = view.currentRenderPassDescriptor,
      let renderEncoder =
        commandBuffer.makeRenderCommandEncoder(
          descriptor: descriptor) else {
        return
    }
      
      uniforms.viewMatrix = float4x4(translation: [0, 0, -3]).inverse
      
    renderEncoder.setVertexBytes(&uniforms,
                                   length: MemoryLayout<Uniforms>.stride, index: 1)
    renderEncoder.setRenderPipelineState(pipelineState)
    renderEncoder.setVertexBuffer(sphere.buffer, offset: 0, index: 0)
    
    renderEncoder.setTriangleFillMode(.lines)
    renderEncoder.drawPrimitives(type: .triangle, vertexStart: 0, vertexCount: sphere.vertexCount)

    renderEncoder.endEncoding()
    guard let drawable = view.currentDrawable else {
      return
    }
    commandBuffer.present(drawable)
    commandBuffer.commit()
  }
}
